# Work around a bug in how ArcGIS hosts the Python interpreter. This
# bug, acknowledged by ESRI, prevents the Python extension modules
# from working properly if they are run more than once.

import sys, _GeoEcoArcGISHelper
_GeoEcoArcGISHelper.LoadModulesFromCache()
try:

    # Initialize the geoprocessor using MGET.

    from GeoEco.ArcGIS import GeoprocessorManager
    GeoprocessorManager.InitializeGeoprocessor()
    gp = GeoprocessorManager.GetWrappedGeoprocessor()

    # Initialize MGET's logging and do all actions inside a try/except
    # block, so we can report exception details when verbose logging is
    # enabled.

    from GeoEco.Logging import Logger
    Logger.Initialize(activateArcGISLogging=True)

    try:
        # Parse the tool arguments.

        inputFile = gp.GetParameterAsText(0)
        variable = gp.GetParameterAsText(1)
        outputFeatureClass = gp.GetParameterAsText(2)

        # Open the HDF file and validate that we can process the requested
        # variable.

        from GeoEco.Datasets.HDF4 import HDF4SDSCollection

        hdf = HDF4SDSCollection(inputFile, lazyPropertyValues={'Dimensions': u'yx', 'PhysicalDimensions': u'yx', 'PhysicalDimensionsFlipped': (False, False)})

        def _ReadSDS(hdf, sdsName):
            datasets = hdf.QueryDatasets('SDSName = \'%s\'' % sdsName, reportProgress=False)
            if len(datasets) <= 0:
                raise ValueError('HDF file %s does not contain a variable named %s.' % (inputFile, sdsName))
            sds = datasets[0]
            if len(sds.Shape) != 2:
                raise ValueError('The variable %s in HDF file %s has %i dimensions. It must have 2 dimensions to be processed by this tool.' % (sdsName, inputFile, len(sds.Shape)))
            return sds

        variableSDS = _ReadSDS(hdf, variable)
        try:
            longitudeSDS = _ReadSDS(hdf, 'Longitude')
        except:
            longitudeSDS = _ReadSDS(hdf, 'longitude')
        try:
            latitudeSDS = _ReadSDS(hdf, 'Latitude')
        except:
            latitudeSDS = _ReadSDS(hdf, 'latitude')

        if longitudeSDS.Shape != variableSDS.Shape:
            raise ValueError('In HDF file %s, the Longitude and %s variables do not have the same shape. They must have the same shape to be processed by this tool.' % (inputFile, variable))
        if latitudeSDS.Shape != variableSDS.Shape:
            raise ValueError('In HDF file %s, the Longitude and %s variables do not have the same shape. They must have the same shape to be processed by this tool.' % (inputFile, variable))

        # Look up the No Data values and scaling equations and apply
        # them.

        def _ApplyHDFAttributes(sds):
            sdsObj = sds._GetSDS()

            # First try to look up the No Data value that should be
            # used before any scaling equation is applied. Note that
            # NASA GSFC OceanColor uses bad_value_scaled (not
            # bad_value_unscaled) for this value. See
            # http://oceancolor.gsfc.nasa.gov/REPROCESSING/R2009/format/'
            # for more information.

            for attrName in ['_FillValue', 'bad_value_scaled']:
                if hasattr(sdsObj, attrName):
                    if sds.DataType.startswith('float'):
                        sds.SetLazyPropertyValue('UnscaledNoDataValue', int(getattr(sdsObj, attrName)))
                    else:
                        sds.SetLazyPropertyValue('UnscaledNoDataValue', float(getattr(sdsObj, attrName)))
                    break

            # Now, if we found that No Data value and the variable
            # contains integers, try to look up the scaling equation
            # for converting those integers to floats. If we find one,
            # compute the scaled No Data value. (When we say "scaled"
            # here, we mean the float value computed from the
            # integers, not the other way around, as NASA GSFC
            # OceanColor seems to mean in their documentation.)

            if sds.UnscaledNoDataValue is not None and not sds.DataType.startswith('float'):
                import numpy
                for (slopeAttrName, interceptAttrName) in {'slope': 'intercept', 'scale_factor': 'add_offset'}.items():
                    if hasattr(sdsObj, slopeAttrName) and hasattr(sdsObj, interceptAttrName):
                        sds.SetLazyPropertyValue('ScaledDataType', u'float32')
                        sds.SetLazyPropertyValue('ScalingFunction', lambda data: numpy.cast['float32'](float(getattr(sdsObj, slopeAttrName))*data + float(getattr(sdsObj, interceptAttrName))))
                        sds.SetLazyPropertyValue('ScaledNoDataValue', sds.GetLazyPropertyValue('ScalingFunction')(sds.UnscaledNoDataValue))
                        break

        _ApplyHDFAttributes(variableSDS)
        _ApplyHDFAttributes(longitudeSDS)
        _ApplyHDFAttributes(latitudeSDS)

        # Read the data.

        values = variableSDS.Data[:].flatten()
        longitudes = longitudeSDS.Data[:].flatten()
        latitudes = latitudeSDS.Data[:].flatten()

        # Create the empty output feature class and add a Value field.

        Logger.Info('Inserting points into feature class %s representing the %s variable of HDF file %s.' % (outputFeatureClass, variable, inputFile))

        import os
        from GeoEco.Datasets.ArcGIS import ArcGISTable

        gp.CreateFeatureClass_management(os.path.dirname(outputFeatureClass), os.path.basename(outputFeatureClass), 'POINT', None, 'DISABLED', 'DISABLED', "GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]")
        points = ArcGISTable(outputFeatureClass)
        points.AddField('Value', variableSDS.DataType)

        # Open an insert cursor and add all of the points. If a point
        # has the No Data value, do not add it.

        ogr = ArcGISTable._ogr()

        cur = points.OpenInsertCursor()
        try:
            for i in range(len(values)):
                if values[i] != variableSDS.NoDataValue and longitudes[i] != longitudeSDS.NoDataValue and latitudes[i] != latitudeSDS.NoDataValue:
                    point = ogr.Geometry(ogr.wkbPoint)
                    point.SetPoint_2D(0, float(longitudes[i]), float(latitudes[i]))
                    cur.SetGeometry(point)
                    if variableSDS.DataType.startswith('float'):
                        cur.SetValue('Value', float(values[i]))
                    else:
                        cur.SetValue('Value', int(values[i]))
                    cur.InsertRow()
        finally:
            del cur

    # If we got an exception, log it.

    except:
        Logger.LogExceptionAsError()

# Second part of the workaround of the ArcGIS bug.

finally:
    _GeoEcoArcGISHelper.AddModulesToCache(sys.modules.keys())
